import fire
from .namespaces import MakeNamespace, CacheNamespace, InitNamespace


class HomaCommand:
    def __init__(self):
        self.make = MakeNamespace()
        self.cache = CacheNamespace()
        self.init = InitNamespace()


def main():
    fire.Fire(HomaCommand)



import torch
from .ActivationFunction import ActivationFunction


class NReLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        sigma = x.std(unbiased=False)
        a = torch.randn_like(x) * sigma
        return torch.where(x >= 0, x + a, 0)



import torch
import random


def replace_activations(module, needle: torch.nn.Module, candidates: list):
    for name, module in module.named_children():
        if isinstance(module, needle):
            factory = random.choice(candidates)
            new_module = factory()
            setattr(module, name, new_module)
        else:
            replace_activations(module, needle, candidates)



import torch
from .Swin import Swin
from .StochasticClassifier import StochasticClassifier
from .utils import replace_activations


class StochasticSwin(Swin, StochasticClassifier):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        replace_activations(self.network, torch.nn.GELU, self._activation_pool)



import torch
from ..activations import (
    SGELU,
    LaLU,
    CaLU,
    TripleStateSwish,
    GeneralizedSwish,
    ExponentialSwish,
    SigmoidDerivative,
    Gish,
    Logish,
    LogLogish,
    ExpExpish,
    SelfArctan,
    ParametricLogish,
    Phish,
    Suish,
    TSReLU,
    TBSReLU,
    LogSigmoid,
    ShiftedReLU,
    RReLU,
    SoftsignRReLU,
    RandomizedSlopedReLU,
    NoisyReLU,
    SineReLU,
    MinSin,
)


class StochasticClassifier:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._activation_pool = [
            torch.nn.ELU,
            torch.nn.PReLU,
            torch.nn.ReLU,
            torch.nn.ReLU6,
            torch.nn.RReLU,
            torch.nn.SELU,
            torch.nn.CELU,
            torch.nn.GELU,
            torch.nn.SiLU,
            torch.nn.Mish,
            SGELU,
            LaLU,
            CaLU,
            TripleStateSwish,
            GeneralizedSwish,
            ExponentialSwish,
            SigmoidDerivative,
            Gish,
            Logish,
            LogLogish,
            ExpExpish,
            SelfArctan,
            ParametricLogish,
            Phish,
            Suish,
            TSReLU,
            TBSReLU,
            LogSigmoid,
            ShiftedReLU,
            RReLU,
            SoftsignRReLU,
            RandomizedSlopedReLU,
            NoisyReLU,
            SineReLU,
            MinSin,
        ]



import torch
from .ActivationFunction import ActivationFunction


class TBSReLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = 1 - torch.exp(-x)
        b = 1 + torch.exp(-x)
        c = a / b
        return x * torch.tanh(c)



from .HasLabels import HasLabels
from .HasLogits import HasLogits
from .HasProbabilities import HasProbabilities
from .ReportsAccuracy import ReportsAccuracy
from .ReportsMetrics import ReportsMetrics
from .Trainable import Trainable



import torch
from .Classifier import Classifier
from .concerns import Trainable, ReportsMetrics
from .modules import SwinModule
from ..device import get_device


class Swin(Classifier, Trainable, ReportsMetrics):
    def __init__(self, num_classes: int, lr: float = 0.0001):
        super().__init__()
        self.network = SwinModule(num_classes=num_classes).to(get_device())
        self.optimizer = torch.optim.AdamW(self.network.parameters(), lr=lr)
        self.criterion = torch.nn.CrossEntropyLoss()



from .GaussianReLU import GaussianReLU


class GaLU(GaussianReLU):
    def __init__(
        self,
        channels: int | None = None,
        max_input: float = 1.0,
    ):
        self.hats = [(2.0, 2.0), (1.0, 1.0), (3.0, 1.0)]
        super().__init__(self.hats, channels=channels, max_input=max_input)



import torch


class ExpExpish(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.exp(-torch.exp(-x))



import torch


class HasLabels:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def predict(self, x: torch.Tensor):
        return torch.argmax(self.logits(x), dim=1)

    @torch.no_grad()
    def predict_(self, x: torch.Tensor):
        return torch.argmax(self.logits(x), dim=1)



from .MexicanReLU import MexicanReLU


class WideMeLU(MexicanReLU):
    def __init__(self, channels: int | None = None, max_input: float = 1.0):
        self.hats = [
            (2.0, 2.0),
            (1.0, 1.0),
            (3.0, 1.0),
            (0.5, 0.5),
            (1.5, 0.5),
            (2.5, 0.5),
            (3.5, 0.5),
        ]
        super().__init__(self.hats, channels=channels, max_input=max_input)



from .Classifier import Classifier
from .Model import Model
from .Resnet import Resnet
from .StochasticResnet import StochasticResnet
from .Swin import Swin
from .StochasticSwin import StochasticSwin



from pathlib import Path
import ast


class InitNamespace:
    def class_(self):
        path = Path(".")
        init_file = path / "__init__.py"
        init_file.write_text("")
        for file in path.iterdir():
            if file.name == "__init__.py" or file.suffix != ".py":
                continue
            tree = ast.parse(file.read_text())
            classes = [
                node.name
                for node in tree.body
                if isinstance(node, ast.ClassDef) and not node.name.startswith("_")
            ]
            functions = [
                node.name
                for node in tree.body
                if isinstance(node, ast.FunctionDef) and not node.name.startswith("_")
            ]
            if not (classes or functions):
                continue
            module = file.stem
            lines = [
                f"from .{module} import {name}\n" for name in (*classes, *functions)
            ]
            with init_file.open("a") as f:
                f.writelines(lines)
            print(f"Processed {file}: classes={classes}, functions={functions}")



import torch


class ReportsLogits:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def logits(self, x: torch.Tensor) -> torch.Tensor:
        batch_size = x.shape[0]
        logits = torch.zeros((batch_size, self.num_classes))
        for model in self.models:
            logits += model(x)
        return logits

    @torch.no_grad()
    def logits_(self, *args, **kwargs):
        return self.logits(*args, **kwargs)



import torch
from .Resnet import Resnet
from .StochasticClassifier import StochasticClassifier
from .utils import replace_activations


class StochasticResnet(Resnet, StochasticClassifier):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        replace_activations(self.network, torch.nn.ReLU, self._activation_pool)



class ReportsSize:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def size(self):
        return len(self.models)

    @property
    def length(self):
        return len(self.models)
